;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "elfeed" "4.0.1"
  "An Atom/RSS feed reader."
  '((emacs  "28.1")
    (compat "31"))
  :url "https://github.com/emacs-elfeed/elfeed"
  :commit "90470c3ff7cf215771f31e8c5e4b08fe6ad8f65b"
  :revdesc "4.0.1-0-g90470c3ff7cf"
  :keywords '("network" "comm" "hypermedia")
  :authors '(("Christopher Wellons" . "wellons@nullprogram.com"))
  :maintainers '(("Karthik Chikmagalur" . "karthik.chikmagalur@gmail.com")
                 ("Ihor Radchenko" . "yantar92@posteo.net")
                 ("Daniel Mendler" . "mail@daniel-mendler.de")))
